﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace Chapter4
{
    public interface IService
    {
        string DoWork(int parm1);

        string DoLessWork();

        string DoMoreWork(int parm1, string parm2);

        int GetANumber();
    }

    public class ClientProxy<T>
        where T : class
    {
        public ClientProxy()
        {
        }

        public TResult CallInterface<TResult>(Expression<Func<T, TResult>> op)
        {
            var exp = op.Body as MethodCallExpression;
            var methodName = exp.Method.Name;
            var methodInfo = exp.Method;
            var allParameters = from element in exp.Arguments
                                select processArgument(element);
            Console.WriteLine($"Calling {methodName}");

            foreach (var parm in allParameters)
                Console.WriteLine($"\tParameter type = {parm.ParmType}, Value = {parm.ParmValue}");
            return default(TResult);
        }

        private (Type ParmType, object ParmValue) processArgument(Expression element)
        {
            object argument = default(object);
            LambdaExpression expression = Expression.Lambda(
                Expression.Convert(element, element.Type));
            Type parmType = expression.ReturnType;
            argument = expression.Compile().DynamicInvoke();
            return (parmType, argument);
        }
    }

    public class SourceContact
    {
        public string Name { get; set; }
        public string Address { get; set; }

        public int NotFound { get; set; }
    }

    public class DestinationContact
    {
        public string Name { get; set; }
        public string Address { get; set; }

        public int Missing { get; set; }

        public override string ToString()
        {
            return string.Format("Name: {0}, Address: {1}, Missing: {2}", Name, Address, Missing);
        }
    }

    public class Converter<TSource, TDest>
        where TDest : new()
    {

        private Func<TSource, TDest> converter;

        private void createConverterIfNeeded()
        {
            if (converter == null)
            {
                var source = Expression.Parameter(typeof(TSource), "source");
                var dest = Expression.Variable(typeof(TDest), "dest");

                var assignments = from srcProp in typeof(TSource).GetProperties(
                                      BindingFlags.Public | BindingFlags.Instance)
                                  where srcProp.CanRead
                                  let destProp = typeof(TDest).GetProperty(
                                      srcProp.Name,
                                      BindingFlags.Public | BindingFlags.Instance)
                                  where (destProp != null) && (destProp.CanWrite)
                                  select Expression.Assign(
                                      Expression.Property(dest, destProp),
                                      Expression.Property(source, srcProp));

                // put together the body:
                var body = new List<Expression>();
                body.Add(Expression.Assign(dest,
                    Expression.New(typeof(TDest))));
                body.AddRange(assignments);
                body.Add(dest);

                var expr =
                    Expression.Lambda<Func<TSource, TDest>>(
                        Expression.Block(
                        new[] { dest }, // expression parameters
                        body.ToArray() // body
                        ),
                        source  // lambda expression
                    );

                var func = expr.Compile();
                converter = func;
            }
        }

        public TDest ConvertFrom(TSource source)
        {
            createConverterIfNeeded();
            return converter(source);
        }

#if false
        // Not legal C#, explanation only
        TDest ConvertFromImaginary(TSource source)
        {
            TDest destination = new TDest();
            foreach (var prop in sharedProperties)
                destination.prop = source.prop;
            return destination;
        }
#endif
    }


    public class Item43
    {
        public static void Snippet01()
        {
            var client = new ClientProxy<IService>();
            var result = client.CallInterface<string>(
                srver => srver.DoWork(172));
        }

        public static void Snippet02()
        {
            var client = new ClientProxy<IService>();
            client.CallInterface(srver => srver.DoWork(
                client.CallInterface(srv => srv.GetANumber())));
        }

        public static void Snippet03()
        {
            SourceContact source = new SourceContact { Address = "addr", Name = "Name", NotFound = 25 };

            var converter = new Converter<SourceContact,
                DestinationContact>();
            DestinationContact dest2 = converter.ConvertFrom(source);
        }
    }
}


